"""This module provides the UpdateManager class for managing updates for a program."""

import os
import requests
import tempfile
import shutil
import pickle
import subprocess
import sys
from packaging.version import Version
from pyupgrader.utilities import helper, hashing


class DBSumError(Exception):
    """This exception is raised when there is an error in comparing values from the databases."""
    pass


class GetFilesError(Exception):
    """This exception is raised when there is an error in retrieving file paths from the cloud."""
    pass


class DownloadFilesError(Exception):
    """This exception is raised when there is an error in downloading files from the cloud."""
    pass


class NoUpdateError(Exception):
    """This exception is raised when there is no files downloaded during an update."""
    pass


class UpdateManager:
    """
    Class for managing updates for a program.

    Attributes:
    - url: str
        URL to the .pyupgrader folder
    - project_path: str
        Path to the project folder (Not the .pyupgrader folder)
    
    Methods:
    - check_update() -> dict
        Compare cloud and local version and return a dict with the results
    - db_sum() -> DBSummary
        Return a DBSummary object using the cloud and local hash databases
    - get_files(updated_only: bool = False) -> list
        Retrieves a list of files from the cloud database.
    - download_files(save_path: str = "", required: bool = False) -> str
        Download files to save_path, if save_path is empty, create a temp folder, return the save_path
    - update(file_dir: str = "") -> str
        Start the application update process.
    """

    def __init__(self, url: str, project_path: str):
        """
        Initialize the UpdateManager class.

        Args:
        - url: str
            URL to the .pyupgrader folder
        - project_path: str
            Path to the project folder (Not the .pyupgrader folder)
        """
        self._url = url.rstrip('/')
        self._project_path = project_path.rstrip('/')
        self._pyupgrader_path = os.path.join(self._project_path, '.pyupgrader')
        self._config_path = os.path.join(self._pyupgrader_path, 'config.yaml')
        self._local_hash_db_path = None  # Set in _validate_attributes

        self._config_man = helper.Config()
        self._web_man = None  # Set in _validate_attributes

        self._validate_attributes()

    @property
    def url(self) -> str:
        """
        Get the URL to the .pyupgrader folder.

        Returns:
        - str: The URL to the .pyupgrader folder.
        """
        return self._url

    @url.setter
    def url(self, value: str) -> None:
        """
        Set the URL to the .pyupgrader folder.

        Args:
        - value: str
            The URL to the .pyupgrader folder.
        """
        self._url = value.rstrip('/')  # Remove trailing slash
        self._web_man = helper.Web(self._url)
        self._validate_attributes()

    @property
    def project_path(self) -> str:
        """
        Get the path to the project folder (Not the .pyupgrader folder).

        Returns:
        - str: The path to the project folder.
        """
        return self._project_path

    @project_path.setter
    def project_path(self, value) -> None:
        """
        Set the path to the project folder (Not the .pyupgrader folder).

        Args:
        - value: str
            The path to the project folder.
        """
        self._project_path = value
        self._pyupgrader_path = os.path.join(self._project_path, '.pyupgrader')
        self._config_path = os.path.join(self._pyupgrader_path, 'config.yaml')
        self._local_hash_db_path = None  # Set in _validate_attributes
        self._validate_attributes()
    
    def _validate_attributes(self) -> None:
        """
        Validate and set attributes of the class.
        """
        if not os.path.exists(self._project_path):
            raise FileNotFoundError(self._project_path)
        try:
            requests.get(self._url)
        except requests.exceptions.ConnectionError:
            raise requests.exceptions.ConnectionError(self._url)
        if not os.path.exists(self._pyupgrader_path):
            raise FileNotFoundError(self._pyupgrader_path)
        if not os.path.exists(self._config_path):
            raise FileNotFoundError(self._config_path)
        
        config_data = self._config_man.load_yaml(self._config_path)
        self._local_hash_db_path = os.path.join(self._pyupgrader_path, config_data['hash_db'])
        self._web_man = helper.Web(self._url)

        if not os.path.exists(self._local_hash_db_path):
            raise FileNotFoundError(self._local_hash_db_path)

    def check_update(self) -> dict:
        """
        Compare cloud and local version and return a dict with the results.

        Returns:
        - dict: A dictionary with the following keys:
            - has_update (bool): True if there is an update available, False otherwise.
            - description (str): The description of the update.
            - web_version (str): The version number from the cloud.
            - local_version (str): The version number from the local configuration.
        """
        web_config = self._web_man.get_config()
        local_config = self._config_man.load_yaml(self._config_path)

        web_version = Version(web_config['version'])
        local_version = Version(local_config['version'])

        if web_version > local_version:
            has_update, description = True, web_config['description']
        else:
            has_update, description = False, local_config['description']
        
        return {'has_update': has_update, 'description': description, 'web_version': str(web_version), 'local_version': str(local_version)}
    
    def db_sum(self) -> hashing.DBSummary:
        """
        Return a DBSummary object using the cloud and local hash databases.

        Returns:
        - hashing.DBSummary: A DBSummary object.
        """
        tmp_path = ""
        try:
            tmp_path = tempfile.mkdtemp()
            cloud_hash_db_path = self._web_man.download_hash_db(os.path.join(tmp_path, 'cloud_hashes.db'))

            return hashing.compare_databases(self._local_hash_db_path, cloud_hash_db_path)
        except Exception as error:
            raise DBSumError(error)
        finally:
            if os.path.exists(tmp_path):
                shutil.rmtree(tmp_path)
    
    def get_files(self, updated_only: bool = False) -> list:
        """
        Retrieves a list of files from the cloud database.
        Note that this function does not return files that have been deleted from the cloud.

        Args:
        - updated_only (bool, optional): If True, only returns files that have been updated. 
            Defaults to False.

        Returns:
        - list: A list of file paths.

        Raises:
        - Exception: If an error occurs during the retrieval process.
        """
        db_temp_path = ""
        cloud_db = None
        try:
            db_temp_path = tempfile.mkdtemp()

            cloud_hash_db_path = self._web_man.download_hash_db(os.path.join(db_temp_path, 'cloud_hashes.db'))
            cloud_db = hashing.HashDB(cloud_hash_db_path)
            compare_db = self.db_sum()

            files = None

            if updated_only:
                bad_files = [path for path, _, _ in compare_db.bad_files]
                files = compare_db.unique_files_cloud_db + bad_files
            else:
                files = [path for path in cloud_db.get_file_paths()]

            return files
        except Exception as error:
            raise GetFilesError(error)
        finally:
            if cloud_db is not None:
                cloud_db.close()
            if db_temp_path:
                shutil.rmtree(db_temp_path)
    
    def download_files(self, save_path: str = "", required: bool = False) -> str:
        """
        Download cloud files and return the path where the files are saved.

        Args:
        - save_path: str, optional
            The path to save the downloaded files. If not provided, a temporary folder will be created.
        - required: bool, optional
            If True, only download files that have changed or have been added.

        Returns:
        - str: The path where the files are saved.

        Raises:
        - Exception: If an error occurs during the download process.
        """
        try:
            if not save_path:
                save_path = tempfile.mkdtemp()
            
            files_to_download = None
            
            if required:
                files_to_download = self.get_files(updated_only=True)
            else:
                files_to_download = self.get_files()

            # Download all files in db and copy structure
            base_url = self._url.split(".pyupgrader")[0]
            for file_path in files_to_download:
                download_url = base_url + file_path

                # Create save path
                relative_path = os.path.dirname(file_path)
                save_folder = os.path.join(save_path, relative_path)
                os.makedirs(save_folder, exist_ok=True)
                save_file = os.path.join(save_folder, os.path.basename(file_path))

                self._web_man.download(download_url, save_file)
            
            return save_path
        except Exception as error:
            raise DownloadFilesError(error)
    
    def update(self, file_dir: str = "") -> str:
            """
            Start the application update process.
            A lock file will be created in the .pyupgrader folder.
            This file is used by file_updater.py to determine when to start updating.
            Remove the lock file to start the update process,
            main application should call sys.exit() immediately after removing the lock file.

            Args:
            - file_dir (str, optional):
                The directory where temporary files will be stored. If not provided, a temporary directory will be created.

            Returns:
            - str: The path to the lock file.

            Raises:
            - NoUpdateError: If there are no files to update. Set required_only to False in the cloud config to update anyway.
            """
            # init values
            cloud_config = self._web_man.get_config()
            db_summary = self.db_sum()
            download_files = False
            if not file_dir:
                file_dir = tempfile.mkdtemp()
                download_files = True

            # Create temp folder in file_dir for holding update settings
            tmp_setting_dir = tempfile.mkdtemp(dir=file_dir)
            cloud_config_path = os.path.join(tmp_setting_dir, 'config.yaml')
            cloud_hash_db_path = os.path.join(tmp_setting_dir, 'hashes.db')

            # Populate settings folder
            self._web_man.download_hash_db(cloud_hash_db_path)
            self._config_man.write_yaml(cloud_config_path, cloud_config)

            update_details = {
                'update': [],
                'delete': [file_path for file_path in db_summary.unique_files_local_db],
                'project_path': self._project_path,
                'startup_path': os.path.join(self._project_path, cloud_config['startup_path']),
                'cloud_config_path': cloud_config_path,
                'cloud_hash_db_path': cloud_hash_db_path,
            }

            # Set the 'update' value and download files as needed
            if not cloud_config['required_only']:
                if download_files:
                    self.download_files(file_dir, required=False)
                update_details['update'] = self.get_files(updated_only=False)
            else:
                if download_files:
                    self.download_files(file_dir, required=True)
                update_details['update'] = [file_path for file_path in db_summary.unique_files_cloud_db] + [file_path for file_path, _, _ in db_summary.bad_files]
            
            if cloud_config['required_only'] and not update_details['update'] and not update_details['delete']:
                shutil.rmtree(file_dir)
                raise NoUpdateError('No files to update. | If you wish to update anyway, set required_only to False in the cloud config.')

            # save actions to pickle file
            action_pkl = os.path.join(tmp_setting_dir, 'actions.pkl')
            with open(action_pkl, 'wb') as file:
                pickle.dump(update_details, file)
            
            # create lock file
            lock_file = os.path.join(self._pyupgrader_path, 'lock')
            with open(lock_file, 'w') as file:
                file.write('')
            
            # start file_updater.py using subprocessing
            command = [sys.executable, os.path.join(os.path.dirname(__file__), 'utilities', 'file_updater.py'), '-p', file_dir, '-a', action_pkl, '-l', lock_file]
            if cloud_config['cleanup']:
                command.append('-c')
            subprocess.Popen(command)

            return lock_file



"""
This module contains utility functions and classes for PyUpgrader.

Functions:
- normalize_paths(paths: Union[str, List[str]]) -> List[str]: Replace backslashes with forward slashes in a path or a list of paths.

Classes:
- Config: Helper class for managing configuration files.
- Web: Class for managing web requests.
"""

import yaml
import requests
import pkg_resources
import yaml
from typing import List, Union


def normalize_paths(paths: Union[str, List[str]]) -> List[str]:
    """
    Replace backslashes with forward slashes in a path or a list of paths.

    Args:
        paths (Union[str, List[str]]): A path or a list of paths.

    Returns:
        List[str]: A list of paths with backslashes replaced by forward slashes.

    Raises:
        TypeError: If the input is not a string or a list of strings.
    """
    if isinstance(paths, str):
        return paths.replace('\\', '/')
    elif isinstance(paths, list):
        return [path.replace('\\', '/') for path in paths]
    else:
        raise TypeError("Input must be a string or a list of strings")


class Config:
    """
    Config helper class

    Attributes:
    default_config_path: str
        Path to the default config file
    comments_path: str
        Path to the comments file

    Methods:
    load_yaml(path: str) -> dict
        Load a yaml file at path
    loads_yaml(yaml_string: str) -> dict
        Load a yaml from a string
    write_yaml(path: str, data: dict) -> None
        Dump data to yaml file at path
    display_info() -> None
        Display config values and comments
    """

    def load_yaml_from_package(self, package_name: str, file_path: str) -> dict:
        """
        Load a YAML file from a package.

        Args:
            package_name (str): The name of the package.
            file_path (str): The path to the YAML file inside the package.

        Returns:
            dict: The data loaded from the YAML file.
        """
        file_content = pkg_resources.resource_string(package_name, file_path)
        data = yaml.safe_load(file_content)
        return data

    def load_yaml(self, path: str) -> dict:
        """
        Load a yaml file at path.

        Args:
        - path (str): The path to the yaml file.

        Returns:
        dict: The data loaded from the yaml file.
        """
        with open(path, 'r') as config_file:
            data = yaml.safe_load(config_file)
            is_valid, error = self._valid_config(data)
            if not is_valid:
                raise ValueError(error)
            return data
    
    def loads_yaml(self, yaml_string: str) -> dict:
        """
        Load a yaml from a string.

        Args:
        - yaml_string (str): The yaml string to load.

        Returns:
        dict: The data loaded from the yaml string.
        """
        data = yaml.safe_load(yaml_string)
        is_valid, error = self._valid_config(data)
        if not is_valid:
            raise ValueError(error)
        return data
    
    def write_yaml(self, path: str, data: dict) -> None:
        """
        Dump data to yaml file at path.

        Args:
        - path (str): The path to the yaml file.
        - data (dict): The data to dump to the yaml file.
        """
        with open(path, 'w') as config_file:
            yaml.safe_dump(data, config_file)

    def display_info(self) -> None:
        """
        Display config values and comments.

        Prints the config values and their corresponding comments.
        """
        comments = self.load_yaml_from_package('pyupgrader', 'utilities/comments.yaml')
        config = self.load_yaml_from_package('pyupgrader', 'utilities/default.yaml')

        header = "Config Information"
        print(f"""\n\t{header}\n\t{'-' * len(header)}\n\tAttributes marked as Dynamic can be changed by the user\n""")

        misc_comments = {}

        # Display config values and comments
        for key, value in config.items():
            print(f"{key}: {value}")
            if key in comments:
                print(f"  Comments: {comments[key]}")
            else:
                misc_comments[key] = comments.get(key, "")
            print()

        # Display misc comments
        if misc_comments:
            print("Misc Comments:")
            for key, value in misc_comments.items():
                print(f"{key}: {value}\n")
    
    def _valid_config(self, config: dict) -> (bool, str):
        """
        Validate the config.

        Args:
        - config (dict): The config to validate.

        Returns:
        tuple: A tuple containing a boolean indicating if the config is valid and a string describing the error if it is not valid.
        """
        if 'version' not in config:
            return False, 'Missing "version" attribute'
        if 'description' not in config:
            return False, 'Missing "description" attribute'
        if 'hash_db' not in config:
            return False, 'Missing "hash_db" attribute'
        if 'startup_path' not in config:
            return False, 'Missing "startup_path" attribute'
        if 'required_only' not in config:
            return False, 'Missing "required_only" attribute'
        if 'cleanup' not in config:
            return False, 'Missing "cleanup" attribute'

        return True, ""


class Web:
    """
    Class for managing web requests
    
    Attributes:
    url: str
        URL to the .pyupgrader folder
    
    Methods:
    get_request(url: str) -> requests.Response
        Get a request from the url
    get_config() -> dict
        Get the config file from the url
    download(url_path: str, save_path) -> str
        Download a file from the url_path and save it to save_path
    download_hash_db(save_path: str) -> str
        Download the hash database and save it to save_path
    """
    def __init__(self, url: str):
        """
        Initialize the Web class with the provided URL.
        
        Args:
        - url (str): URL to the .pyupgrader folder
        """
        self._url = url
        self._config_url = self._url + '/config.yaml'
        self._config_man = Config()
    
    def get_request(self, url: str) -> requests.Response:
        """
        Get a request from the specified URL.
        
        Parameters:
        - url (str): URL to send the request to
        
        Returns:
        - requests.Response: The response object from the request
        """
        try:
            response = requests.get(url)
            response.raise_for_status()
        except Exception as e:
            raise requests.ConnectionError(f'Url: "{url}" | {e}')
        
        return response
    
    def get_config(self) -> dict:
        """
        Get the config file from the URL.
        
        Returns:
        - dict: The parsed config file as a dictionary
        """
        response = self.get_request(self._config_url)
        return self._config_man.loads_yaml(response.text)
    
    def download(self, url_path: str, save_path) -> str:
        """
        Download a file from the specified URL path and save it to the specified save path.
        
        Args:
        - url_path (str): URL path of the file to download
        - save_path: Path to save the downloaded file
        
        Returns:
        - str: The save path of the downloaded file
        """
        response = self.get_request(url_path)

        with open(save_path, 'wb') as f:
            f.write(response.content)
        
        return save_path
    
    def download_hash_db(self, save_path: str) -> str:
        """
        Download the hash database and save it to the specified save path.
        
        Args:
        - save_path (str): Path to save the hash database file
        
        Returns:
        - str: The save path of the downloaded hash database file
        """
        config = self.get_config()
        return self.download(self._url + '/' + config['hash_db'], save_path)



"""
Hasher module for PyUpgrader.

This module provides classes and functions for hashing files, creating hash databases, and comparing hash databases.
Modified version of https://github.com/Trogiken/random-projects/blob/master/python/tools/DataIntegrityChecker/SCRIPT.py

Classes:
- DBSummary: Dataclass for database summary.
- HashDB: A class that provides methods to interact with a hash database.
- Hasher: A class that provides methods for hashing files and creating hash databases.

Functions:
- compare_databases(local_db_path: str, cloud_db_path: str) -> DBSummary: Compare two hash databases and return a summary of the differences.

Exceptions:
- HashingError: Exception raised for errors in the hashing process.
"""

import hashlib
import os
import sqlite3
import time
import re
from multiprocessing import Pool
from dataclasses import dataclass
from pyupgrader.utilities import helper


@dataclass
class DBSummary:
    """
    Dataclass for database summary

    Attributes
    ----------
    unique_files_local_db: list
        List of files unique to the local database
    unique_files_cloud_db: list
        List of files unique to the cloud database
    ok_files: list
        List of in-common files that have the same hash
    bad_files: list
        List of in-common files that have different hashes
    """
    unique_files_local_db: list
    unique_files_cloud_db: list
    ok_files: list
    bad_files: list


class HashingError(Exception):
    """Exception raised for errors in the hashing process."""
    pass


def compare_databases(local_db_path: str, cloud_db_path: str) -> DBSummary:
    """
    Compare two hash databases and return a summary of the differences.

    - local_db_path (str): The file path of the local hash database.
    - cloud_db_path (str): The file path of the cloud hash database.

    Returns:
        DBSummary: An object containing the summary of the differences between the two databases.
    """
    connection1 = sqlite3.connect(local_db_path)
    cursor1 = connection1.cursor()

    connection2 = sqlite3.connect(cloud_db_path)
    cursor2 = connection2.cursor()

    cursor1.execute('SELECT file_path, calculated_hash FROM hashes')
    local_db_files = {row[0]: row[1] for row in cursor1.fetchall()}

    cursor2.execute('SELECT file_path, calculated_hash FROM hashes')
    cloud_db_files = {row[0]: row[1] for row in cursor2.fetchall()}

    common_files = set(local_db_files.keys()) & set(cloud_db_files.keys())
    unique_files_local_db = list(set(local_db_files.keys()) - set(cloud_db_files.keys()))
    unique_files_cloud_db = list(set(cloud_db_files.keys()) - set(local_db_files.keys()))

    ok_files = [
        (file_path, local_db_files[file_path])
        for file_path in common_files
        if local_db_files[file_path] == cloud_db_files[file_path]
    ]

    bad_files = [
        (file_path, local_db_files[file_path], cloud_db_files[file_path])
        for file_path in common_files
        if local_db_files[file_path] != cloud_db_files[file_path]
    ]
    
    connection1.close()
    connection2.close()

    return DBSummary(
        unique_files_local_db=unique_files_local_db,
        unique_files_cloud_db=unique_files_cloud_db,
        ok_files=ok_files,
        bad_files=bad_files
    )


class HashDB:
    """
    A class that provides methods to interact with a hash database.

    Attributes:
        db_path (str): The path to the hash database.

    Methods:
        get_file_paths() -> str: Generator that yields file paths from the database.
        get_file_hash(file_path: str) -> str: Returns the hash of a file in the database.
        open() -> None: Opens the database connection.
        close() -> None: Closes the database connection.
    """

    def __init__(self, db_path: str):
        """
        Initializes a new instance of the HashDB class.

        Args:
        - db_path (str): The path to the hash database.
        """
        self.db_path = db_path
        self.connection = None
        self.cursor = None
        self.open()
    
    def get_file_paths(self) -> str:
        """
        Generator that yields file paths from the database.
        """
        self.cursor.execute('SELECT file_path FROM hashes')
        for row in self.cursor.fetchall():
            yield row[0]
    
    def get_file_hash(self, file_path: str) -> str:
        """
        Returns the hash of a file in the database.

        Args:
        - file_path (str): The path of the file.

        Returns:
        - str: The hash of the file.
        """
        self.cursor.execute('SELECT calculated_hash FROM hashes WHERE file_path = ?', (file_path,))
        return self.cursor.fetchone()[0]
    
    def open(self) -> None:
        """
        Opens the database connection.
        """
        self.connection = sqlite3.connect(self.db_path)
        self.cursor = self.connection.cursor()

    def close(self) -> None:
        """
        Closes the database connection.
        """
        self.connection.close()


class Hasher:
    """
    A class that provides methods for hashing files and creating hash databases.

    Attributes:
    - project_name: str
        The name of the project directory (Not the full path)

    Methods:
    - create_hash(self, file_path: str) -> (str, str): Creates a hash from file bytes using the chunk method and returns the relative file path and hash as a string.
    - create_hash_db(self, hash_dir_path: str, db_save_path: str, exclude_paths=[], exclude_patterns=[]) -> str: Creates a hash database from a directory path and saves it to a file path. Returns the file path.
    """

    def __init__(self, project_name: str):
        """
        Initialize the Hasher class.

        Args:
        - project_name: str
            The name of the project directory (Not the full path)
        """
        self.project_name = project_name

    def create_hash(self, file_path: str) -> (str, str):
        """
        Create a hash from file bytes using the chunk method and return the relative file path and hash as a string.

        Args:
        - file_path: str
            The path of the file to be hashed.

        Returns:
        - Tuple[str, str]: The relative file path and hash as a string.
        """
        try:
            chunk_size = 4096
            file_size = os.path.getsize(file_path)

            if file_size > 1_000_000_000:  # If file size around 1 Gb or larger
                chunk_size = 8192

            hasher = hashlib.sha256()

            with open(file_path, 'rb') as file:
                while True:
                    chunk = file.read(chunk_size)
                    if not chunk:
                        break
                    hasher.update(chunk)

            relative_file_path = file_path.split(self.project_name)[-1].lstrip('/')
            file_hash = hasher.hexdigest()

            return relative_file_path, file_hash
        except Exception as error:
            raise HashingError(f"Error hashing file '{file_path}' | {error}")

    def create_hash_db(self, hash_dir_path: str, db_save_path: str, exclude_paths=[], exclude_patterns=[]) -> str:
        """
        Create a hash database from a directory path and save it to a file path. Return the save file path.

        Args:
        - hash_dir_path: str
            The path of the directory to create the hash database from.
        - db_save_path: str
            The path to save the hash database file.
        - exclude_paths: List[str], optional
            A list of paths to exclude from the hash database creation. Default is an empty list.
        - exclude_patterns: List[str], optional
            A list of patterns to exclude from the hash database creation. Default is an empty list.

        Returns:
        - str: The file path of the saved hash database.
        """
        if os.path.exists(db_save_path):
            os.remove(db_save_path)

        exclude_paths = helper.normalize_paths(exclude_paths)

        # separate files and directories from exclude_paths
        exclude_file_paths = [path for path in exclude_paths if os.path.isfile(path)]
        exclude_dir_paths = [path for path in exclude_paths if os.path.isdir(path)]

        connection = sqlite3.connect(db_save_path)
        cursor = connection.cursor()

        # Create table for hashes
        cursor.execute('''CREATE TABLE IF NOT EXISTS hashes (
                            file_path TEXT PRIMARY KEY,
                            calculated_hash TEXT
                        )''')

        # Batch size for parameterized queries
        max_time_per_batch = 3  # seconds
        batch_data = []

        # Create a pool, default number of processes is the number of cores on the machine
        with Pool() as pool:
            start_time = time.time()  # Start timer

            for root, dirs, files in os.walk(hash_dir_path):
                if exclude_dir_paths:
                    # If the root directory is in the exclude directories
                    if any(exclude_dir_path in helper.normalize_paths(root) for exclude_dir_path in exclude_dir_paths):
                        dirs[:] = []  # Skip subdirectories
                        continue

                if exclude_patterns:
                    # if any directory in the root matches a pattern, skip it
                    if any(re.search(pattern, helper.normalize_paths(root)) for pattern in exclude_patterns):
                        dirs[:] = []  # Skip subdirectories
                        continue

                file_paths = helper.normalize_paths([os.path.join(root, file) for file in files])  # Get full file paths

                if exclude_file_paths:
                    for path in exclude_file_paths:
                        if path in file_paths:
                            file_paths.remove(path)

                if exclude_patterns:
                    # if any file in the root matches a pattern, skip it
                    file_paths = [
                        path
                        for path in file_paths
                        if not any(re.search(pattern, path) for pattern in exclude_patterns)
                    ]

                results = pool.map(self.create_hash, file_paths)  # Use workers to create hashes
                batch_data.extend(results)

                elapsed_time = time.time() - start_time
                if elapsed_time >= max_time_per_batch and batch_data:  # If the max time per batch has been reached and there are files to be inserted
                    cursor.executemany('INSERT OR REPLACE INTO hashes (file_path, calculated_hash) VALUES (?, ?)', batch_data)
                    batch_data = []
                    start_time = time.time()

            if batch_data:  # If there are any remaining files to be inserted
                cursor.executemany('INSERT OR REPLACE INTO hashes (file_path, calculated_hash) VALUES (?, ?)', batch_data)

        connection.commit()
        connection.close()

        return db_save_path



"""PyUpgrader CLI"""

import argparse
import os
import sys
import pyupgrader.utilities as util
from pyupgrader.utilities.build import BuildError


def cli():
    """PyUpgrader CLI

    This function is the entry point for the PyUpgrader command-line interface (CLI).
    It parses the command-line arguments, validates the project folder path,
    and builds the project using the specified options.

    Args:
        -p, --project: Absolute path to project folder
        -no_env: Exclude environment directories
        -no_hidden: Exclude hidden files and directories
        -patterns: Exclude files and directories using regex patterns
        -e, --exclude: Absolute paths for excluded files and directories

    Raises:
        BuildError: If an error occurs during the build process.
    """
    parser = argparse.ArgumentParser(description='PyUpgrader CLI')
    parser.add_argument('-p', '--project', help="Absolute path to project folder", required=True)
    parser.add_argument('-no_env', help="Exclude environment directories", action='store_true')
    parser.add_argument('-no_hidden', help="Exclude hidden files and directories", action='store_true')
    parser.add_argument('-patterns', help="Exclude files and directories using regex patterns", nargs='+', default=[])
    parser.add_argument('-e', '--exclude', help="Absolute paths for excluded files and directories", nargs='+', default=[])
    args = parser.parse_args()
    
    if not os.path.exists(args.project):
        print(f'Folder "{args.project}" does not exist')
        sys.exit(1)
    
    try:
        builder = util.Builder(project_path=args.project, exclude_envs=args.no_env, exclude_hidden=args.no_hidden, exclude_patterns=args.patterns, exclude_paths=args.exclude)
        builder.build()
    except Exception as error:
        raise BuildError(error)



"""
This module provides the functionality to build a project into a pyupgrader project.
It is intended to be used by the command line tool.

It defines the following classes:
- BuildError: Raised when there is an error building a project.
- FolderCreationError: Raised when there is an error creating a folder.
- ConfigError: Raised when there is an error with the config file.
- HashDBError: Raised when there is an error with the hash database.
- PathError: Raised when there is an error with a path.
- Builder: Builds a project into a pyupgrader project.
"""
import os
import shutil
from pyupgrader.utilities import helper, hashing


class BuildError(Exception):
    """Raised when there is an error building a project"""
    pass


class FolderCreationError(Exception):
    """Raised when there is an error creating a folder"""
    pass


class ConfigError(Exception):
    """Raised when there is an error with the config file"""
    pass


class HashDBError(Exception):
    """Raised when there is an error with the hash database"""
    pass


class PathError(Exception):
    """Raised when there is an error with a path"""
    pass


class Builder:
    """
    Builds a project into a pyupgrader project

    Attributes:
    - project_path: Path to the project folder.
    - exclude_envs: Whether to exclude common virtual environment folders.
    - exclude_hidden: Whether to exclude hidden files and folders.
    - exclude_patterns: List of patterns to exclude from the hash database.
    - exclude_paths: List of absolute paths to exclude from the hash database.

    Methods:
    - build(): Builds the project into a pyupgrader project.
    - _validate_paths(): Validates and sets the paths.
    - _create_pyupgrader_folder(): Creates the .pyupgrader folder.
    - _create_config_file(): Creates the config file.
    - _create_hash_db(): Creates the hash database.
    """

    def __init__(self, project_path: str, exclude_envs: bool = False, exclude_hidden: bool = False, exclude_patterns: list = [], exclude_paths: list = []):
        self.project_path = project_path
        self.exclude_envs = exclude_envs
        self.exclude_hidden = exclude_hidden
        self.exclude_patterns = exclude_patterns
        self.exclude_paths = exclude_paths

        self.env_names = [
            '.venv',
            'venv',
            'env',
            '.env',
            '.virtualenv',
            'virtualenv',
            'conda',
            '.conda',
            'condaenv',
            '.condaenv',
            'pipenv',
            '.pipenv',
            'poetry',
            '.poetry',
            'pyenv',
            '.pyenv'
        ]

        self._pyudpdate_folder = None
        self._config_path = None
        self._hash_db_path = None

    def build(self):
        """Builds a project into a pyupgrader project"""
        self._validate_paths()

        print('Building project...\n')

        try:
            self._create_pyupgrader_folder()
        except Exception as error:
            raise FolderCreationError(f'Failed to create .pyupgrader folder | {error}')
        
        try:
            self._create_config_file()
        except Exception as error:
            raise ConfigError(f'Failed to create config file | {error}')
        
        try:
            self._create_hash_db()
        except Exception as error:
            raise HashDBError(f'Failed to create hash database | {error}')

        print('\nDone!')
        print(f'Project built at "{self._pyudpdate_folder}"')
    
    def _validate_paths(self):
        """Validates and set paths"""
        if self.project_path is None:
            raise BuildError('Folder path not set')
        if self.exclude_paths is None:
            raise BuildError('Exclude paths not set')

        if not os.path.exists(self.project_path):
            raise FileNotFoundError(f'Folder "{self.project_path}" does not exist')
        if self.project_path in self.exclude_paths:
            raise PathError(f'Folder path cannot be excluded')
        
        # Remove trailing slashes
        if self.project_path.endswith('/'):
            self.project_path = self.project_path.rstrip('/')
        if self.project_path.endswith('\\'):
            self.project_path = self.project_path.rstrip('\\')
        for i, path in enumerate(self.exclude_paths):
            if path.endswith('/'):
                self.exclude_paths[i] = path.rstrip('/')
            if path.endswith('\\'):
                self.exclude_paths[i] = path.rstrip('\\')
        
        self._pyudpdate_folder = os.path.join(self.project_path, '.pyupgrader')
        self._config_path = os.path.join(self._pyudpdate_folder, 'config.yaml')
        self._hash_db_path = os.path.join(self._pyudpdate_folder, 'hashes.db')
    
    def _create_pyupgrader_folder(self):
        """Creates the .pyupgrader folder"""
        if os.path.exists(self._pyudpdate_folder):
            print(f'Folder "{self._pyudpdate_folder}" already exists')
            print('Deleting folder')
            shutil.rmtree(self._pyudpdate_folder)

        print(f'Creating folder at "{self._pyudpdate_folder}"')
        os.mkdir(self._pyudpdate_folder)

    def _create_config_file(self):
        """Creates the config file"""
        print(f'Creating config file at "{self._config_path}"')
        config = helper.Config()

        default_data = config.load_yaml_from_package('pyupgrader', 'utilities/default.yaml')
        default_data['hash_db'] = os.path.basename(self._hash_db_path)
        config.write_yaml(self._config_path, default_data)
        
        try:
            config.load_yaml(self._config_path)
        except ValueError as error:
            raise ConfigError(f'Failed to validate config file | {error}')
    
    def _create_hash_db(self):
        """Creates the hash database"""
        print(f'Creating hash database at "{self._hash_db_path}"')
        hasher = hashing.Hasher(project_name=os.path.basename(self.project_path))

        self.exclude_paths.append(self._pyudpdate_folder)
        self.exclude_patterns.append(r'.*/__pycache__/.*')

        if self.exclude_hidden:
            self.exclude_patterns.append(r'.*/\..*')
        if self.exclude_envs:
            self.exclude_paths += [os.path.join(self.project_path, path) for path in self.env_names]

        hasher.create_hash_db(self.project_path, self._hash_db_path, self.exclude_paths, self.exclude_patterns)



"""This module is a utility for updating files."""

import argparse
import os
import pickle
import shutil
import subprocess
import sys
from time import sleep
from pyupgrader.utilities import hashing


def main():
    # Create the argument parser
    parser = argparse.ArgumentParser(description='Update Utility')

    parser.add_argument('-p', '--path', help='Path to downloaded files', required=True)
    parser.add_argument('-a', '--action', help='Path to the action file', required=True)
    parser.add_argument('-l', '--lock', help='Path to the lock file', required=True)
    parser.add_argument('-c', '--clean', help='Clean the downloads path', action='store_true')
    args = parser.parse_args()

    # Wait for the lock file to be removed
    while os.path.exists(args.lock):
        sleep(.5)
    sleep(1)

    with open(args.action, 'rb') as f:
        update_details = pickle.load(f)

        update_files = update_details['update']
        del_files = update_details['delete']
        project_path = update_details['project_path']
        startup_path = update_details['startup_path']
        cloud_config_path = update_details['cloud_config_path']
        cloud_hash_db_path = update_details['cloud_hash_db_path']
    
    # Replace config and hash db
    if os.path.exists(cloud_config_path):
        source = cloud_config_path
        destination = os.path.join(project_path, '.pyupgrader', os.path.basename(cloud_config_path))
        os.remove(destination)
        shutil.copy(source, destination)
    else:
        raise FileNotFoundError(f"Cloud config not found at '{cloud_config_path}'")
    
    if os.path.exists(cloud_hash_db_path):
        source = cloud_hash_db_path
        destination = os.path.join(project_path, '.pyupgrader', os.path.basename(cloud_hash_db_path))
        os.remove(destination)
        shutil.copy(source, destination)
    else:
        raise FileNotFoundError(f"Cloud hash db not found at '{cloud_hash_db_path}'")
    
    # Update the files
    for file in update_files:
        source = os.path.join(args.path, file)
        destination = os.path.join(project_path, file)
        if os.path.exists(destination):
            os.remove(destination)
        shutil.copy(source, destination)
    
    for file in del_files:
        destination = os.path.join(project_path, file)
        if os.path.exists(destination):
            os.remove(destination)
    
    if args.clean:
        shutil.rmtree(args.path)

    subprocess.Popen([sys.executable, startup_path])

    sys.exit(0)


if __name__ == '__main__':
    try:
        main()
    except Exception as e:
        with open(os.path.join(os.path.dirname(__file__), 'update_crash.txt'), 'w') as f:
            f.write(str(e))
        raise e



from setuptools import setup

with open('requirements.txt') as f:
    requirements = f.read().splitlines()


setup(
    name='pyupgrader',
    version='1.0.0b1',
    author='Noah Blaszak',
    author_email='70231827+Trogiken@users.noreply.github.com',
    description='Python library that allows the version of a program to be updated.',
    packages=['pyupgrader', 'pyupgrader.utilities'],
    package_data={'pyupgrader.utilities': ['default.yaml', 'comments.yaml']},
    install_requires=requirements,
    entry_points={
        'console_scripts': [
            'pyupgrader=pyupgrader.main_cli:cli'
        ]
    },
)
